/*
  simple - a formal system demonstrating cyclic values
  Copyright (C) 1996  Scott Draves <spot@cs.cmu.edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundationeither version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this programif not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "time.h"
#include "runtime.h"
#include "names.h"


int sum_shorts_hand(short *start, short *stop, int sum) {
  while (start != stop) {
    sum += *start++;
  }
  return sum;
}

int scan_int_hand(short *start, short *stop,
		  int *start1, int *stop1, int sum) {
  while (start != stop) {
    *start1++ = sum;
    sum += *start++;
  }
  return sum;
}

void add_shorts_hand(short *start, short *stop,
		     short *start1, short *stop1,
		     short *start2, short *stop2) {
  while (start != stop) {
    *start2++ = *start++ + *start1++;
  }
}

void add_shorts_hand2(short *start, short *stop, int stride,
		      short *start1, short *stop1, int stride1,
		      short *start2, short *stop2, int stride2) {
  while (start != stop) {
    *start2 = *start + *start1;
    start2 += stride2;
    start += stride;
    start1 += stride1;
  }
}

void add_shorts_hand2_inter(short *start, short *stop,
			    short *start2, short *stop2) {
  while (start != stop) {
    *start2++ = start[0] + start[1];
    start += 2;
  }
}

void add_shorts_hand4(short *start, short *stop,
		    short *start1, short *stop1,
		    short *start2, short *stop2) {
  while (start != stop) {
    start2[0] = start[0] + start1[0];
    start2[1] = start[1] + start1[1];
    start2[2] = start[2] + start1[2];
    start2[3] = start[3] + start1[3];
    start2 += 4;
    start += 4;
    start1 += 4;
  }
}

void add_10_shorts_hand(short *start, short *stop,
			short *start2, short *stop2) {
  while (start != stop) {
    *start2++ = *start++ + 10;
  }
}

void add_10_shorts_hand4(short *start, short *stop,
			short *start2, short *stop2) {
  while (start != stop) {
    start2[0] = start[0] + 10;
    start2[1] = start[1] + 10;
    start2[2] = start[2] + 10;
    start2[3] = start[3] + 10;
    start2 += 4;
    start += 4;
  }
}

void filter2_hand(short *start, short *stop,
		  short *start1, short *stop1) {
  while (start != stop) {
    *start1 = start[0] + start[1];
    start++;
    start1++;
  }
}

void filter5_hand(short *start, short *stop,
		  short *start1, short *stop1) {
  int i;
  int t;
  while (start != stop) {
    t = 0;
    for (i = 0; i < 5; i++)
      t  += start[i];
    *start1 = t;
    start++;
    start1++;
  }
}

void lut1_m(int *lut, unsigned char *start, unsigned char *stop,
	    short *start1, short *stop1)
{
  while (start != stop)
    *start1++ = lut[*start++];
}

void lut1_m4(int *lut, unsigned char *start, unsigned char *stop,
	     short *start1, short *stop1)
{
  while (start != stop) {
    start1[0] = lut[start[0]];
    start1[1] = lut[start[1]];
    start1[2] = lut[start[2]];
    start1[3] = lut[start[3]];
    start1 += 4;
    start += 4;
  }
}

void lut1_m42(int *lut, unsigned char *start, unsigned char *stop,
	      short *start1, short *stop1)
{
  int *s0 = (int *) start;
  int *e0 = (int *) stop;
  int *s1 = (int *) start1;
  while (s0 != e0) {
    int w = *s0;
    s1[0] = lut[w&0xff] | (lut[(w>>8)&0xff] << 16);
    s1[1] = lut[(w>>16)&0xff] | (lut[(w>>24)&0xff] << 16);
    s0 += 1;
    s1 += 2;
  }
}

void wavtab1_m(int *lut, int prev, short *start, short *stop,
	       short *start1, short *stop1)
{
  while (start != stop)
    prev = *start1++ = lut[((*start++) + (prev>>5))&1023];
}

void fm1_m(int *lut, int phase, short *start, short *stop)
{
  while (start != stop) {
    *start++ = lut[phase>>8];
    phase += 256;
    phase = phase & ((1024*256)-1);
  }
}

#if 0
on x86 M[a+(b<<2)] is faster than M[(a+b)<<2]
#endif

void fm1_m2(int *lut, int phase, short *start, short *stop)
{
  int *s0 = (int *) start;
  int *e0 = (int *) stop;
  int a, b;
  while (s0 != e0) {
    a = lut[phase>>8];
    phase += 256;
    phase = phase & ((1024*256)-1);
    b = lut[phase>>8];
    phase += 256;
    phase = phase & ((1024*256)-1);
    *s0++ = a | (b << 16);
  }
}

void fm3_m(int *lut, int phase0, int phase1, short *start, short *stop)
{
  while (start != stop) {
    phase0 += 256;
    if (phase0 >= (1024*256))
      phase0 = phase0 - (1024*256);

    phase1 += lut[phase0>>8]/256 + 256;
    if (phase1 >= (1024*256))
      phase1 = phase1 - (1024*256);
    if (phase1 < 0)
      phase1 = phase1 + (1024*256);
    *start++ = lut[phase1>>8];

  }
}

main(int argc, char **argv)
{
  int n;
  int s2;
  short u[2100]; /* 2^10 words, also used as wavtab */
  short v[1000];
  short w[2000];
  int iters = 1000;
  int v_len = 1000;
  int ddddum;
  int nsamps;
  int vv = v_len / 2;

  word_t start = ((word_t) v>>2);
  word_t start1 = ((word_t) w>>2);
  word_t start2 = ((word_t) u>>2);
  word_t dummy = ((int)&ddddum)>>2;

  s2 = 0;
  begin_timer();
  for (n = 0; n < iters; n++) {
    s2 += sum_shorts_hand(v, v+v_len, 0);
  }
  printf("m sum_shorts %.3g\n", end_timer());

  s2 = 0;
  begin_timer();
  for (n = 0; n < iters; n++) {
    s2 += sum_shorts(start, start+vv, 0, 0, 0);
  }
  printf("a sum_shorts %.3g\n", end_timer());

  s2 = 0;
  begin_timer();
  for (n = 0; n < iters; n++) {
    s2 += scan_int_hand(v, v+v_len, (int *) w, (int *) w+v_len, 0);
  }
  printf("m scan_int %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    scan_ints(start, start+vv, start1, start1 + v_len,
	      0, dummy, 0, dummy, 0);
  }
  printf("a scan_int %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_shorts_hand(v, v+v_len, w, w+v_len, u, u+v_len);
  }
  printf("m add_shorts %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_shorts(start, start+vv, start1, start1+vv, start2, start2+vv,
	       0, dummy, 0, dummy, 0, dummy, 0);
  }
  printf("a add_shorts %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_shorts_hand4(v, v+v_len, w, w+v_len, u, u+v_len);
  }
  printf("m add_shorts4 %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_10_shorts_hand(v, v+v_len, w, w+v_len);
  }
  printf("m add_10_shorts %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_10_shorts_hand4(v, v+v_len, w, w+v_len);
  }
  printf("m add_10_shorts4 %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_10_shorts(start, start+vv, start1, start1+vv,
		  0, dummy, 0, dummy, 0);
  }
  printf("a add_10_shorts %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    copy_filter2(0, start, start+vv, start1, start1+vv,
		 0, dummy, 0, dummy, 0);
  }
  printf("a copy_filter2 %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    filter2_hand(v, v+v_len, w, w+v_len);
  }
  printf("m filter2 %.3g\n", end_timer());
  
  begin_timer();
  for (n = 0; n < iters; n++) {
    filter5(start, start+vv, 0, 0, 0, 0, 0,
	    start1, start1+vv,
	    0, dummy, 0, dummy, 0);
  }
  printf("a filter5 %.3g\n", end_timer());
  
  begin_timer();
  for (n = 0; n < iters; n++) {
    filter5_hand(v, v+v_len, w, w+v_len);
  }
  printf("m filter5 %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_shorts_hand2(v, v+2*v_len, 2, v+1, v+1+2*v_len, 2, u, u+v_len, 1);
  }
  printf("m add_shorts2 inter %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_shorts_hand2(v, v+v_len, 1, w, w+v_len, 1, u, u+v_len, 1);
  }
  printf("m add_shorts2 seq %.3g\n", end_timer());

  begin_timer();
  for (n = 0; n < iters; n++) {
    add_shorts_hand2_inter(v, v+2*v_len, u, u+v_len);
  }
  printf("m add_shorts2_inter %.3g\n", end_timer());

  for (n = 0; n < 256; n++)
    u[n] = n;
  for (n = 0; n < v_len; n++)
    v[n] = random();

  lut1(start2, start, start+v_len/4, start1, start1+v_len/2,
       0, dummy, 0, dummy, 0);

  begin_timer();
  for (n = 0; n < iters; n++) {
    lut1(start2, start, start+v_len/4, start1, start1+v_len/2,
	 0, dummy, 0, dummy, 0);
  }
  printf("a lut1 %.3g\n", end_timer());

  lut1_m((int *) u, (unsigned char *) v,
	 (unsigned char *) (v+v_len/2), w, w+v_len);

  begin_timer();
  for (n = 0; n < iters; n++) {
    lut1_m((int *) u, (unsigned char *) v,
	   (unsigned char *) (v+v_len/2), w, w+v_len);
	 
  }
  printf("m lut1 %.3g\n", end_timer());
  
  lut1_m4((int *) u, (unsigned char *) v,
	  (unsigned char *) (v+v_len/2), w, w+v_len);

  begin_timer();
  for (n = 0; n < iters; n++) {
    lut1_m4((int *) u, (unsigned char *) v,
	    (unsigned char *) (v+v_len/2), w, w+v_len);
	 
  }
  printf("m lut1 4 %.3g\n", end_timer());
  

  lut1_m42((int *) u, (unsigned char *) v,
	   (unsigned char *) (v+v_len/2), w, w+v_len);

  begin_timer();
  for (n = 0; n < iters; n++) {
    lut1_m42((int *) u, (unsigned char *) v,
	     (unsigned char *) (v+v_len/2), w, w+v_len);
	 
  }
  printf("m lut1 42 %.3g\n", end_timer());

  for (n = 0; n < 1023; n++)
    ((int*)u)[n] = n;
  for (n = 0; n < v_len; n++)
    v[n] = random();

  wavtab1(start2, 0, start, start+(v_len/2),
	  start1, start1+(v_len/2),
	  0, dummy, 0, dummy, 0);

  begin_timer();
  for (n = 0; n < iters; n++) {
    wavtab1(start2, 0, start, start+(v_len/2),
	    start1, start1+(v_len/2),
	    0, dummy, 0, dummy, 0);
  }
  printf("a wavtab1 %.3g\n", end_timer());

  wavtab1_m((int *) u, 0, v, (v+v_len), w, w+v_len);
  begin_timer();
  for (n = 0; n < iters; n++) {
    wavtab1_m((int *) u, 0, v, (v+v_len), w, w+v_len);
	 
  }
  printf("m wavtab1 %.3g\n", end_timer());

  fm1(start2, 0, start, start+(v_len/2), 0, dummy, 0);
  begin_timer();
  for (n = 0; n < iters; n++) {
    fm1(start2, 0, start, start+(v_len/2), 0, dummy, 0);
  }
  printf("a fm1 %.3g\n", end_timer());

  fm1_m((int *) u, 0, v, (v+v_len));
  begin_timer();
  for (n = 0; n < iters; n++) {
    fm1_m((int *) u, 0, v, (v+v_len));
	 
  }
  printf("m fm1 %.3g\n", end_timer());

  fm1_m2((int *) u, 0, v, (v+v_len));
  begin_timer();
  for (n = 0; n < iters; n++) {
    fm1_m2((int *) u, 0, v, (v+v_len));
	 
  }
  printf("m fm12 %.3g\n", end_timer());

  fm2(start2, 0, start, start+(v_len/2), 0, dummy, 0);
  begin_timer();
  for (n = 0; n < iters; n++) {
    fm2(start2, 0, start, start+(v_len/2), 0, dummy, 0);
  }
  printf("a fm2 %.3g\n", end_timer());

  fm3(start2, 0, 0, start, start+(v_len/2), 0, dummy, 0);
  begin_timer();
  for (n = 0; n < iters; n++) {
    fm3(start2, 0, 0, start, start+(v_len/2), 0, dummy, 0);
  }
  printf("a fm3 %.3g\n", end_timer());

  fm3_m((int *) u, 0, 0, v, (v+v_len));
  begin_timer();
  for (n = 0; n < iters; n++) {
    fm3_m((int *) u, 0, 0, v, (v+v_len));
	 
  }
  printf("m fm3 %.3g\n", end_timer());

}
