(*
  simple - a formal system demonstrating cyclic values
  Copyright (C) 1996  Scott Draves <spot@cs.cmu.edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundationeither version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this programif not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)



functor Eval (structure Syntax : SYNTAX) (* : EVAL *) =
struct
    structure Syntax = Syntax
    open Syntax

    (* structure M_Control = Control (type ans = Syntax.M) *)

type Env = (Name -> M)

exception unbound of string

exception bad_bases of Prim * myint * myint

val debug_out:(TextIO.outstream option ref) = ref NONE

fun DO s = print s
(*
fun DO s = case (!debug_out) of
            NONE => ()
          | SOME(port) => (TextIO.output(port, s);
                           TextIO.flushOut(port))
*)
fun minus (x:myint) = 0w0 - x

val minint = Word32.<<(0w1,0w31)

fun mydiv (x, y) = if (0w0 = y) then (print "div by zero"; raise Div)
                   else if (x < minint)
                   then x div y
                   else (minus ((minus x) div y))
                   

fun mymod (x, y) = if (0w0 = y) then (print "mod by zero"; raise Div)
                   else if (x < minint)
                   then x mod y
                   else (minus ((minus x) mod y))
                   
                      


fun empty_env v = (raise unbound(v))

fun extend (m_dynamic(exp_var(v))) c env =
     (fn v' => (if (v = v') then c else (env v')))
  | extend (m_pair(p0,p1)) (m_pair(v0,v1)) env =
    (extend p0 v0 (extend p1 v1 env))
  | extend p v env = (print("cannot bind " ^ (unparse_m p) ^ " to " ^ (unparse_m v) ^ "\n");
                      raise Match)

val genvar = Token.gen_name "v"
val genlabel = Token.gen_name "L"



(*
fun match (m_dynamic _) (m_dynamic _) = true
  | match (m_cyclic(b0,_,r0)) (m_cyclic(b1,_,r1)) = (b0=b1) andalso (r0=r1)
  | match (m_static a0) (m_static a1) = (a0 = a1)
  | match (m_pair(l0,r0)) (m_pair(l1,r1)) = (match l0 l1) andalso (match r0 r1)
  | match _ _ = false
*)

fun match m0 m1 =
    let val bindings = ref []
        fun bind x y = case (List.find (fn (l,r) => x=l) (!bindings)) of
                         SOME (l,r) => r=y
                       | NONE => case (List.find (fn (l,r) => x=r) (!bindings)) of
                                   SOME (l,r) => l=x
                                 | NONE => ((bindings := (x,y)::(!bindings));
					    true)
        fun mat (m_dynamic d0) (m_dynamic d1) = bind d0 d1
          | mat (m_cyclic(b0,q0,r0)) (m_cyclic(b1,q1,r1)) =
                (b0=b1) andalso (r0=r1) andalso (bind q0 q1)
          | mat (m_static a0) (m_static a1) = (a0 = a1)
          | mat (m_pair(l0,r0)) (m_pair(l1,r1)) = (mat l0 l1) andalso (mat r0 r1)
          | mat _ _ = false
    in mat m0 m1
    end
 
  

local
  fun ver (ks,km) ({store,formal,label,residual}::tl) =
                if (match ks store) andalso (match km formal)
                then SOME(label,residual) else (ver (ks,km) tl)
    | ver _ _ = NONE

  fun remo (ks, km) ((hd as {store,formal,label,residual})::tl) acc =  
               if (match ks store) andalso (match km formal)
               then (remo (ks,km) tl acc)
               else (remo (ks,km) tl (hd::acc))
    | remo k _ acc = acc
in
  fun remove_version k versions = (versions := remo k (!versions) [])
  fun add_version (ks,km) (lab,m) versions = (versions := ({store=ks,formal=km,label=lab,
                                                                    residual=m} :: 
                                                               (remo (ks,km) (!versions) [])))
  fun find_version k (ref versions) = (ver k versions)
end

fun gen_simple_pattern () = m_dynamic(exp_var(genvar()))

(* generalize
fun make_formal v = case v of
      (m_dynamic v) => gen_simple_pattern()
    | (m_cyclic(b,q,r)) => m_cyclic(b,exp_var(genvar()),r)
    | (m_static v) => (m_static v)
    | (m_pair(l,r)) => m_pair(make_formal l, make_formal r)
 *)

(* generalize *)
fun make_formal v =
    let val bindings = ref []
        fun bind (d:Exp) = case (List.find (fn (l,r)=>l=d) (!bindings)) of
                     NONE => let val v' = exp_var(genvar())
                             in bindings := (d,v')::(!bindings);
                                v'
                             end
                     | SOME(_,r) => (r:Exp)
        fun mf m = case m of
         (m_dynamic d) => m_dynamic(bind d)
       | (m_cyclic(b,q,r)) => m_cyclic(b,(bind q),r)
       | (m_static v) => (m_static v)
       | (m_pair(l,r)) => m_pair(mf l, mf r)
    in
      mf v
    end




fun assign v v' = case (v,v') of
     ((m_dynamic(exp_var v)), v') => 
         (SOME (exp_assign(v, (case (reify_dynamic_part v') of SOME(e) => e | _ => raise Match))))
   | ((m_pair(l,r)),(m_pair(l',r'))) =>
              (case ((assign l l'), (assign r r')) of
                (NONE,NONE) => NONE
              | (SOME x, NONE) => SOME x
              | (NONE, SOME x) => SOME x
              | (SOME x, SOME y) => SOME(exp_comma(x,y)))
   | ((m_cyclic(_,(exp_var v),_)), (m_cyclic(_,v',_)))  =>
               SOME (exp_assign(v, v'))
   | (v, m_effect(e, v')) => (case assign v v' of
                                 NONE => SOME e
                               | SOME x => SOME(exp_comma(e, x)))
   | _ => NONE

fun assign0 v v' = case (assign v v') of SOME x => x
		                       | NONE => exp_atom(atom_unit)

fun par_assign v v' =
    let val tv = make_formal v
    in exp_comma(assign0 tv v', assign0 v tv)
    end


fun all_static (m_static _) = true
  | all_static (m_pair(m0,m1)) = all_static(m0) andalso all_static(m1)
  | all_static _ = false

(* clearly the next two fns should be merged *)

fun find_fun n (def_fn(hd as (name,_,_,_))::tl)  = if n=name then hd else (find_fun n tl)
  | find_fun n (_ :: rest) = (find_fun n rest)
  | find_fun n _ = (print ("no function named " ^ n); raise Match)


fun find_val n (def_val(name,def)::tl)  = if n=name then def else (find_val n tl)
  | find_val n (_ :: rest) = (find_val n rest)
  | find_val n _ = (print (n ^ " unbound"); raise Match)


val trace_eval = ref false
val trace_let = ref false
val trace_apply = ref false


fun same_exp (e0, e1) prog = (e0 = e1)

(* not right but close enough for now *)
fun has_dynamic_if prog env store exp =
   let val D = (has_dynamic_if prog env store)
       val V = (eval prog env store)
   in
   case exp of
         (exp_atom _) => false
       | (exp_var _) => false
       | (exp_app _) => false
       | (exp_if (e0, e1, e2)) => let val (v0,s1) = (V e0)
              in (case v0 of m_dynamic(_) => true | _ => false)
              end
       | (exp_prim(_,es)) => let fun here (e::es) = (D e) orelse (here es)
                                   | here _ = false
                             in (here es) end
       | (exp_let{inline, pattern, exp, body}) => 
                     let val (v',s1) = (V exp) in
                     has_dynamic_if prog (extend pattern v' env) store body
                   end
       | (exp_and(x,y)) => (D x) orelse (D y)  (* no XXX *)
       | (exp_or(x,y)) => (D x) orelse (D y)   (* no XXX *)
       | (exp_loop{index, init, done, next, exit}) => true
   end
 

and eval (prog : Program) env store exp =
   let val V = (eval prog env) in 
   (* (DO("fda "^(unparse_m store)^"\n"));
   (if true then DO ("eval " ^ (unparse exp) ^ "\n") else ()); *)
   case exp of
          (exp_atom c) => ((m_static c), store)
       |  (exp_var v) => (((env v), store) handle (unbound _) => 
                          (V store (find_val v prog)))
       |  (exp_app {label, arg, memoize}) => let val (v0,s1) = V store arg
                                             in apply memoize prog label v0 s1
                                             end
       |  (exp_if(e0, e1, e2)) => let val (p,s1) = V store e0 in
             (case p of 
             (m_static (atom_bool v)) => if v then (V s1 e1) else (V s1 e2) (* XXX lose effects *)
           | (m_dynamic d0) => 
               ((print("warning: dynamic conditional on "^(unparse d0)^"\n"));
               let val (b1,s2) = V s1 e1
                   val (b2,s3) = V s1 e2    (* assert s2 = s3 XXX *)
               in 
                   (m_dynamic(exp_if(d0, (reify b1), (reify b2))), s3)
               end)
           | m => ((print("bad predicate: " ^ (unparse_m m))); raise Match))
           end

       | (exp_quote e) => ((m_dynamic e), store)

       | (exp_let{inline, pattern, exp, body}) => 
               let val (v',s1) = (V store exp)
               in if (!trace_let) then
                     (DO ("let " ^ (Bool.toString inline) ^ " " ^ (unparse_m pattern)
                                      ^ " " ^ (unparse_m v') ^ "\n"))
                     else ();
                  if inline then
                     eval prog (extend pattern v' env) s1 body
                  else
                     let val nv = gen_simple_pattern() 
                         val (b',s2) = eval prog (extend pattern nv env) s1 body
                     in case pattern of m_dynamic(exp_var(v)) =>
                       (m_dynamic(exp_let{inline=true, pattern=nv, exp=reify(v'),
                           body=(reify b')}), s2)
                     end
               end

       | (exp_and(e0,e1)) =>
              let val (v0,s1) = V store e0
              in case v0 of
                 m_static(atom_bool(true)) => (V s1 e1)
               | m_static(atom_bool(false)) => (m_static(atom_bool(false)), s1)
               | m_dynamic(v0) =>
                     let val (v1,s2) = V s1 e1
                       in case v1 of
                         m_static(atom_bool(true)) => (m_dynamic(v0), s2)
                       | m_static(atom_bool(false)) => (m_static(atom_bool(false)), s2)
                       | m_dynamic(v1) => (m_dynamic(exp_and(v0, v1)), s2)
                       | _ => ((print("not a bool "^(unparse_m v1)^"\n"));raise Match)
                       end
              | _ => ((print("not a bool "^(unparse_m v0)^"\n"));raise Match)
              end
       | (exp_or(e0,e1)) => 
             let val (v0, s1) = (V store e0) 
             in case v0 of
                m_static(atom_bool(false)) => (V s1 e1)
              | m_static(atom_bool(true)) => (m_static(atom_bool(true)), s1)
              | m_dynamic(v0) => 
                  let val (v1, s2) = (V s1 e1)
                  in case v1 of
                      m_static(atom_bool(false)) => (m_dynamic(v0), s2)
                    | m_static(atom_bool(true)) => (m_static(atom_bool(true)), s2)
                    | m_dynamic(v1) => (m_dynamic(exp_or(v0, v1)), s2)
                    | _ => ((print("not a bool "^(unparse_m v1)^"\n"));raise Match)
                  end
              | _ => ((print("not a bool "^(unparse_m v0)^"\n"));raise Match)
             end

       (* assume not pure static for now *)
       | (exp_loop{index, init, done, next, exit}) =>
          let val (iv,c1) = V store init
              fun loop0 (v,c1) = let val env1 = extend index v env
                                     val (dv, c2) = eval prog env1 c1 done
                            in case dv of
			        (m_dynamic(e)) =>
				   let val fiv = make_formal v
				       val fic = make_formal c1
				       val env2 = extend index fiv env
				       val (ee,c22) = case (eval prog env2 fic done) 
                                                         of ((m_dynamic x),c) => (x, c)
				       val (nxv,c3) = eval prog env2 c22 next
				       val (nv,c4) = loop1 (fiv,fic) (nxv, c3)
				       val (ev,c5) = eval prog env2 c22 exit
                                   in (exp_comma(par_assign (m_pair(fiv,fic)) (m_pair(v,c1)),
						 exp_comma(exp_while(ee, par_assign (m_pair(fiv,fic))
									 (m_pair(nv,c4))),
							   reify ev)),
				       c5)
                                   end
                               | (m_static(atom_bool p)) =>
                                     if p then
				        let val (v,c3) = eval prog env1 c2 exit
                                        in (reify v, c3) end
				     else
				        let val (v,c3) = eval prog env1 c2 next
                                        in loop0(v, c3) end
                            end
              and loop1 (x0,c0) (v,c1) = let val env1 = extend index v env
	                                     val (dv, c2) = eval prog env1 c1 done
                            in case dv of
			        (m_dynamic(e)) => if (match x0 v) then
                                                  (if (match c2 c0) then
                                                     (v, c2)
                                                  else (print("no match c: "^ (unparse_m c0)^" "^
									    (unparse_m c2));
							     raise Match))
                                                  else (print("no match v: "^ (unparse_m x0)^" "^
									    (unparse_m v));
							     raise Match)
                              | (m_static(atom_bool false)) =>
				        let val (v,c3) = eval prog env1 c2 next
                                        in loop1 (x0,c0) (v, c3) end
                              | _ => (print("exit after enter "^(unparse_m iv)^"\n"); raise Match)
                            end
              val (e,c2) = loop0(iv,c1)
          in
              (m_dynamic(e), c2)
          end

       | (exp_prim (p_set_store, [e])) =>
              let val (v0, s1) = (V store e)
              in
                  (m_static(atom_unit), v0)
              end

       | (exp_prim (p_effect_store, [e])) =>
	    (case (V store e) of (m_dynamic(e1), _) => 
	     (case store of (m_pair((m_dynamic e0), c)) =>
	        (m_static(atom_unit), (m_pair(m_dynamic(exp_comma(e0, e1)),c)))
               | _ => ((print("uuiu"^(unparse_m store))); raise Match) ))

       | (exp_prim (p, es)) =>

        let fun plus z (b, q, r) x = 
             let val rx = (case z of p_plus => r+x
                                   | p_minus => r-x)
                 val r' = mymod(rx, b)
                 val q' = mydiv(rx, b)
             in if 0w0 = q'
                then m_cyclic(b, q, r')
                else m_cyclic(b, exp_prim(z,[exp_atom(atom_int(q')),q]), r')
             end
          fun times (b, q, r) x = if (x > 0w0) 
                 then m_cyclic(b*x, q, r*x)
                 else if (x < 0w0) then m_cyclic(minus(b*x), q, minus(r*x))
                 else m_static(atom_int(0w0))
          fun equals (b, q, r) x = 
                    if mymod(r, b) = mymod(x, b)
                    then m_dynamic(exp_prim(p_equals,[q, exp_atom(atom_int(minus(mydiv(x, b))))]))
                    else m_static(atom_bool(false))

          fun threaded_map (e::es) st = let val (v,s) = V st e
                                            val (vs,ss) = (threaded_map es s)
                                        in ((v :: vs), ss)
                                        end
            | threaded_map _ st = ([],st)
          val (vs, store1) = threaded_map es store
      in
      ((case (p, vs) of

         (p_times, [m_static(atom_int x), m_cyclic(y)]) => (times y x)
       | (p_times, [m_cyclic(y), m_static(atom_int x)]) => (times y x)
       | (p_times, [m_static(atom_int x), m_dynamic(e)]) => m_cyclic(x, e, 0w0)
       | (p_plus, [m_cyclic(y), m_static(atom_int x)]) => (plus p_plus y x)
       | (p_plus, [m_static(atom_int x), m_cyclic(y)]) => (plus p_plus y x)
       | (p_minus, [m_cyclic(y), m_static(atom_int x)]) => (plus p_plus y (minus x))
       | (p_minus, [m_static(atom_int x), m_cyclic(y)]) => (plus p_minus y x)
       | (p_equals, [m_cyclic(y), m_static(atom_int x)]) => (equals y x)
       | (p_equals, [m_static(atom_int x), m_cyclic(y)]) => (equals y x)

       | (p_equals, [m_cyclic(b0, q0, r0),
                     m_cyclic(b1, q1, r1)]) =>
         if (b0 = b1)
         then let val rr0 = mymod(r0, b0)
                  val rr1 = mymod(r1, b0)
                  val rq0 = mydiv(r0, b0)
                  val rq1 = mydiv(r1, b0)
              in if (rr0 = rr1)
                 then m_dynamic(exp_prim(p_equals,
                                if (rq0 = rq1)
                                then [q0, q1]
                                else [q0, exp_prim(p_plus, [q1, exp_atom(atom_int (rq0-rq1))])]))
                 else m_static(atom_bool false)
              end
         else raise bad_bases(p_equals, b0, b1)


       | (p_div, [m_cyclic(b0, q0, r0), m_static(atom_int d)]) =>
           if 0w0 = mymod(b0, d)
           then m_cyclic(mydiv(b0, d), q0, mydiv(r0, d))
           else raise bad_bases(p_div, b0, d)

       | (p_mod, [m_cyclic(b0, q0, r0), m_static(atom_int d)]) =>
           if (b0 = d)
           then m_static(atom_int (mymod(r0, d)))
           else raise bad_bases(p_mod, b0, d)

       | (p_plus, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(x + y))
       | (p_minus, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(x - y))
       | (p_times, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(x * y))
       | (p_equals, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_bool(x = y))
       | (p_div, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(mydiv(x, y)))
       | (p_mod, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(mymod(x, y)))
       | (p_bit_or, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(Word32.orb(x, y)))
       | (p_bit_and, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(Word32.andb(x, y)))
       | (p_bit_shift_left, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(Word32.<<(x,Word31.fromLargeWord(y))))
       | (p_bit_shift_right, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(Word32.>>(x,Word31.fromLargeWord(y))))
       | (p_bit_xor, [m_static (atom_int x), m_static (atom_int y)])
           => m_static(atom_int(Word32.xorb(x, y)))
       | (p_bit_not, [m_static (atom_int x)])
           => m_static(atom_int(Word32.notb(x)))

       | (p_greater, [m_static (atom_int x), m_static (atom_int y)]) => m_static(atom_bool(x > y))
       | (p_less, [m_static (atom_int x), m_static (atom_int y)]) => m_static(atom_bool(x < y))
       | (p_greater_equal, [m_static (atom_int x), m_static (atom_int y)]) => m_static(atom_bool(x >= y))
       | (p_less_equal, [m_static (atom_int x), m_static (atom_int y)]) => m_static(atom_bool(x <= y))

       (* | (p_effect, [v0,v1]) => m_effect(reify v0, v1) *)
       | (p_effect, [v0,v1]) => (case v0 of m_static(atom_unit) => v1 
               | _ => ((print("non unit effect: "^(unparse_m v0)^"\n")); raise Match))

       | (p_aliases, [m_dynamic(e0), m_dynamic(e1)])
           => m_static(atom_bool(same_exp (e0,e1) prog))

       | (p_aliases, [m_cyclic(b0,q0,r0), m_cyclic(b1,q1,r1)])
           => m_static(atom_bool(b0 = b1 andalso (same_exp (q0, q1) prog) andalso r0 = r1))

       | (p_aliases, [m0,m1]) => if true then
                                 m_static(atom_bool(false))
                                 else ((print ("cannot test aliasing between "^
                                          (unparse_m m0)^" "^(unparse_m m1)^"\n"));
                                  raise Match)

       | (p_is_pair, [m_pair(_, _)]) => m_static(atom_bool(true))
       | (p_is_pair, [m_static(v)]) => m_static(atom_bool(false))

       | (p_lift, [m]) => m_dynamic(reify m)
       | (p_is_dynamic, [m_dynamic _]) => (m_static(atom_bool true))
       | (p_is_dynamic, [_]) => (m_static(atom_bool false))


       | (p_debug, [v]) => (print("debug "^(unparse_m v)^"\n");
                            v)

       | (p_pair,[v0,v1]) => m_pair(v0,v1)
       | (p_left,[v]) => (case v of
                           m_pair(v,_) => v   (* loses effects! XXX *)
                         | m_dynamic(e) => m_dynamic(exp_prim(p_left,[e]))
                         | x => (print ((unparse_m x) ^ " not a pair"); raise Match))
       | (p_right,[v]) => (case v of
                           m_pair(_,v) => v   (* loses effects! XXX *)
                         | m_dynamic(e) => m_dynamic(exp_prim(p_right,[e]))
                         | x => (print ((unparse_m x) ^ " not a pair"); raise Match))

       | (p_get_store,[]) => store1


       | (p, es) => m_dynamic(exp_prim(p, (map reify es)))), 

     store1)

     end
   end


and apply force_memo (prog : Program) lab v store =

  ((if (!trace_apply) then DO ("apply " ^ lab ^ " " ^ (unparse_m v) ^ "\n") else ());

  let val (name,formal,body,vs) = (find_fun lab prog)
      val V = eval prog
      val env = (extend formal v empty_env)
      val dyn_if = has_dynamic_if prog env store body
      val all_stat = all_static v
      fun drez lab e = m_dynamic(exp_app{label=lab, arg=e, memoize=false})
  in
    if (not force_memo) andalso ((not dyn_if) orelse all_stat)
    then (V env store body)
    else ((* (DO "warning: memoization\n"); *) 
     case reify_dynamic_part(v) of SOME(darg) =>
       (case (find_version (store,v) vs) of
              NONE => let val olab = genlabel()
                          val fv = make_formal v
                          val fs = make_formal store
                          val nyet = m_dynamic(exp_var("nyet"))
                          val _ = add_version (fs,fv) (olab, (nyet, nyet)) vs
                          val (rez,s2) = (V (extend formal fv empty_env) fs body)
                       in (add_version (fs,fv) (olab, (rez,s2)) vs);
                           (drez olab darg, s2)
                       end
            | SOME(olab,(rez,s2)) => ((drez olab darg), s2))
         | _ => ((print((Bool.toString force_memo)^"; no dynamic part applying "^lab^" to "^(unparse_m v)^"\n"));
                 raise Match))
  end)

  fun get_val n prog = #1(eval prog empty_env (m_static atom_unit) (find_val n prog))

  fun run prog store_name f arg =
      apply false prog f (get_val arg prog) (get_val store_name prog)

  fun clear1 (def_fn(_,_,_,vers)) = vers := []
    | clear1 _ = ()

  fun clear prog = (map clear1 prog; ())

end
