/*
    nitrous - compiler generation for interactive graphics
    Copyright (C) 1995  Scott Draves <spot@cs.cmu.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include "runtime.h"


#include <stdio.h>
#include <string.h>

#define word_t int

word_t *heap;
word_t *freep;

char syms[10000];
char *next_sym;


int make_even(int x, int base) {
  return (x / base) * base;
}

char *intern(char *s)
{
   char *r = next_sym;
   strcpy(next_sym, s);
   next_sym += 1 + strlen(s);
   return r;
}




/* hash XXX */
word_t make_symbol(char *s)
{
   char *t = syms;
   char *r;
   while (t < next_sym) {
      if (!strcmp(t, s)) {
	 r = t;
	 goto done;
      }
      t += 1 + strlen(t);
   }
   r = intern(s);
 done:
   return tag_symbol | (word_t) r;
}

void init_runtime()
{
   heap = (word_t *) malloc(sizeof(word_t) * 1024 * 1024);
   if (heap != heap&mask_data) {
      printf("bad heap location: %x\n", heap);
      exit(1);
   }
   if (untag(make_cprim(init_runtime)) != (word_t)init_runtime) {
      printf("unable to recover cprim: %x\n", init_runtime);
      exit(1);
   }
   freep = heap;
   next_sym = syms;
}

static word_t *the_mark;
void mark_heap() {
   the_mark = freep;
}

void reset_heap() {
   freep = the_mark;
}

word_t print_value(word_t v) {
   switch (v&mask_tag) {
    case tag_symbol:
      printf("%s", v&mask_data);
      break;
    case tag_nil:
      printf("nil");
      break;
    case tag_cons:
      printf("(cons ");
      print_value(car(v));
      printf(" . ");
      print_value(cdr(v));
      printf(")");
      break;
    case tag_string:
      putchar('"');
      puts((char *)untag(v));
      putchar('"');
      break;
    case tag_cprim:
      printf("<cprim %x>", untag(v));
      break;
    default:
      printf("%d", v);
      break;
   }
   return 0;
}

word_t write_string(word_t s) {
   fputs((char *)untag(s), stdout);
   return 0;
}
   

word_t apply_cprim(word_t p, word_t args) {
   if (null_p(args)) {
      return (*(cprim_t)untag(p))();
   } else {
      word_t a1 = car(args);
      word_t args1 = cdr(args);
      if (null_p(args1)) {
	 return (*(cprim_t)untag(p))(a1);
      } else {
	 word_t a2 = car(args1);
	 word_t args2 = cdr(args1);
	 if (null_p(args2)) {
	    return (*(cprim_t)untag(p))(a1, a2);
	 } else {
	    word_t a3 = car(args2);
	    word_t args3 = cdr(args2);
	    if (null_p(args3)) {
	       return (*(cprim_t)untag(p))(a1, a2, a3);
	    } else {
	       printf("too many args to cprim.\n");
	       print_value(p);
	       printf("\n");
	       print_value(args);
	       printf("\n");
	       exit(1);
	    }
	 }
      }
   }
}

word_t error(word_t s, word_t a) {
   printf("fatal error: %s\n", untag(s));
   print_value(a);
   printf("\n");
   exit(1);
}

#if 0

#include <vga.h>

word_t open_vga_fb() {
   vga_modeinfo *modeinfo;
   vga_init();
   vga_setmode(G320x200x256);
   modeinfo = vga_getmodeinfo(G320x200x256);
   return cons((word_t) vga_getgraphmem(),
	       cons(modeinfo->linewidth,
		    cons(320,
			 cons(200, tag_nil))));
}

#else

word_t open_vga_fb() {
   return cons((word_t) malloc(320 * 200),
	       cons(320,
		    cons(320,
			 cons(200, tag_nil))));
}

void vga_setpalvec() {
   /* do nothing */
}

#endif



