/*
  simple - a formal system demonstrating cyclic values
  Copyright (C) 1996  Scott Draves <spot@cs.cmu.edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundationeither version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this programif not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



/* integers have native format, this means that they can overflow into
   another type without raising any kind of error, so watch out! */

#ifndef _runtime_h_
#define _runtime_h_

typedef int word_t;

#ifdef mips
#  define mask_tag    0xFF000000
#  define mask_data   0x10FFFFFF
#  define tag_symbol  0x10000000
#  define tag_nil     0x11000000
#  define tag_cons    0x12000000
#  define tag_ptr     0x13000000
#  define tag_cprim   0x24000000
#  define tag_string  0x15000000
#else
#ifdef hppa
#  define mask_tag    0xFF000000
#  define mask_data   0x40FFFFFF
#  define tag_symbol  0x40000000
#  define tag_nil     0x41000000
#  define tag_cons    0x42000000
#  define tag_ptr     0x43000000
#  define tag_cprim   0x44000000
#  define tag_string  0x45000000
#else
#  define mask_tag    0xFF000000
#  define mask_data   0x00FFFFFF
#  define tag_symbol  0x90000000
#  define tag_nil     0x91000000
#  define tag_cons    0x92000000
#  define tag_ptr     0x93000000
#  define tag_cprim   0x94000000
#  define tag_string  0x95000000
#endif
#endif

#define untag(w) ((w)&mask_data)

#define c_nil   tag_nil
#define c_unit  tag_nil
#define c_true  1
#define c_false 0


#define prim_t inline static word_t
typedef word_t (*cprim_t)();

#if 0
prim_t plus(word_t x, word_t y) { return x + y; }
prim_t minus(word_t x, word_t y) { return x - y; }
prim_t equals(word_t x, word_t y) { return x == y; }
prim_t times(word_t x, word_t y) { return x * y; }
prim_t idiv(word_t x, word_t y) { return x / y; } /* wrong for x < 0 XXX */
prim_t imod(word_t x, word_t y) { return x % y; } /* wrong for x < 0 XXX */
prim_t less(word_t x, word_t y) { return x < y; }
prim_t greater(word_t x, word_t y) { return x > y; }
prim_t less_equal(word_t x, word_t y) { return x <= y; }
prim_t greater_equal(word_t x, word_t y) { return x >= y; }
#define bit_shift_left(x,y) ((x)<<(y))
#define bit_shift_right(x,y) ((x)>>(y))
#define bit_or(x,y) ((x)|(y))
#define bit_and(x,y) ((x)&(y))
prim_t bit_not(word_t x) {return ~x; }
prim_t zero_p(word_t x) {return 0 == x; }
prim_t eq_p(word_t x, word_t y) {return x == y; }
prim_t bool_or(word_t x, word_t y) {return x | y; }
prim_t bool_and(word_t x, word_t y) {return x & y; }
prim_t not(word_t x) {return !x; }
prim_t identity(word_t x) {return x;}
#else
#define plus(x,y) ((x)+(y))
#define minus(x,y) ((x)-(y))
#define times(x,y) ((x)*(y))
#define equals(x,y) ((x)==(y))
#define imod(x,y) ((x)%(y))
#define idiv(x,y) ((x)/(y))
#define less(x,y) ((x)<(y))
#define greater(x,y) ((x)>(y))
#define less_equal(x,y) ((x)<=(y))
#define greater_equal(x,y) ((x)>=(y))
prim_t bit_shift_left(word_t x, word_t y) {return x << y; }
prim_t bit_shift_right(word_t x, word_t y) {return x >> y; }
prim_t bit_or(word_t x, word_t y) {return x | y; }
prim_t bit_and(word_t x, word_t y) {return x & y; }
#define bit_not(x) (~(x))
#define zero_p(x) (0==x)
#define eq_p(x,y) ((x)==(y))
#define bool_or(x,y) ((x)|(y))
#define bool_and(x,y) ((x)&(y))
#define not(x) (!(x))
#define identity(x) (x)
#endif

#define load_word(x)  (*(word_t*)(x<<2))
#define store_word(x,y) (((*(word_t*)(x<<2)) = y),0)
/* prim_t load_word(word_t x) {return (*(word_t*)(x<<2)); } */
/* prim_t store_word(word_t x, word_t y) {(*(word_t*)(x<<2)) = y; return 0; } */
prim_t load_byte(word_t x, word_t y) {return ((char*)x)[y]; }
prim_t store_byte(word_t x, word_t y, word_t z) {((char*)x)[y] = z; return 0; }
prim_t early_equals_p(word_t x, word_t y) {return c_false; }

extern word_t *heap;
extern word_t *freep;

prim_t cons(word_t x, word_t y) {
   word_t r = (word_t) freep;
   freep[0] = x;
   freep[1] = y;
   freep += 2;
   return r|tag_cons;
}
prim_t closure_cons(word_t x, word_t y) {return cons(x, y); }
prim_t car(word_t x) {return ((word_t *) (untag(x)))[0]; }
prim_t cdr(word_t x) {return ((word_t *) (untag(x)))[1]; }
prim_t null_p(word_t x) {return (x&mask_tag) == tag_nil; }
prim_t atom_p(word_t x) {return (x&mask_tag) != tag_cons; }
prim_t pair_p(word_t x) {return (x&mask_tag) == tag_cons; }
prim_t symbol_p(word_t x) {return (x&mask_tag) == tag_symbol; }
prim_t procedure_p(word_t x) {return (x&mask_tag) == tag_cprim; }
#define closure_p procedure_p
prim_t list_p(word_t x) {return null_p(x) || pair_p(x); }

extern word_t apply_cprim(word_t p, word_t args);
extern word_t error(word_t s, word_t a);
extern word_t print_value(word_t v);
extern word_t write_string(word_t s);

extern word_t make_symbol(char *s);
#define make_string(s) (tag_string|(word_t)s)
#define make_cprim(p) (tag_cprim|(word_t)&p)


extern void init_runtime();
extern void mark_heap();
extern void reset_heap();

extern int make_even(int x, int base);

extern word_t open_vga_fb();

#endif
